""" ssm.abcs.loggable
    (boilerplate for logger abstract base)
"""

from ssm import util


class Loggable:
    """
    a mixin that allows subclassers to call things like
    self.logger.debug(..) etc with no hassle
    """

    def __init__(self, name=None, **kwargs):
        if hasattr(self, "logger"):
            return
        logger_name = getattr(self, "logger_name", name or self.__class__.__name__)
        self.logger_name = logger_name
        self.logger = util.get_logger(self.logger_name)



""" ssm.api.manager
"""

import botocore

from .. import abcs

DEFAULT_KMS_ID = "alias/default"  # don't change this


class SecretManager(abcs.Loggable):
    """ """

    def __init__(self, env=None, **kwargs):
        assert env is not None
        self.env = env
        self.logger_name = f"{repr(self.env)}[Secrets]"
        super().__init__(**kwargs)

    def __getitem__(self, name):
        """dictionary compatability"""
        try:
            return self.env.ssm.get_parameter(Name=name, WithDecryption=True)[
                "Parameter"
            ]["Value"]
        except self.env.ssm.exceptions.ParameterNotFound as exc:
            raise KeyError(name)

    def __delitem__(self, name):
        """ex: del secrets['foo']"""
        return self.env.ssm.delete_parameter(Name=name)

    def __setitem__(self, name, value):
        """dictionary compatability"""
        return self.set_secret(name=name, value=value, description=name)

    def set_secret(
        self, name=None, value=None, description=None, kms_id=DEFAULT_KMS_ID
    ):
        """ """
        assert name and value, "cannot set secret without passing name and value"
        description = description or name
        try:
            return self.env.ssm.put_parameter(
                Name=name,
                Value=value,
                Description=description,
                Type="SecureString",
                KeyId=kms_id,
                Tier="Advanced",
                Overwrite=True,
            )
        except botocore.exceptions.ClientError as exc:
            err = "could not set secret `{}` with locals: {}"
            self.logger.critical(err.format(name, locals()))
            raise

    def items(self):
        """dictionary compatability"""
        return {key: self[key] for key in self.keys()}

    @staticmethod
    def _unpack_pager(pages):
        """ """
        out = []
        for p in pages:
            out.extend(p["Parameters"])
        out = {x["Name"]: x["Value"] for x in out}
        return out

    def under(self, path_prefix):
        """ """
        self.logger.debug(f"looking up all SSM keys under {path_prefix}")
        paginator = self.env.ssm.get_paginator("get_parameters_by_path")
        pages = paginator.paginate(
            Path=path_prefix, Recursive=True, WithDecryption=True
        )
        return self._unpack_pager(pages)

    __mod__ = under

    def keys(self, under=None):
        """dictionary compatability"""
        self.logger.debug("looking up all SSM keys")
        paginator = self.env.ssm.get_paginator("describe_parameters")
        pager = paginator.paginate(
            ParameterFilters=[
                dict(
                    Key="Path",
                    Option="Recursive",
                    Values=["/"],
                )
            ]
        )
        return self._unpack_pager(pager).keys()



""" ssm.cli (click boilerplate)
"""

import click

from ssm import util

from . import args, options  # noqa

LOGGER = util.get_logger(__name__)


class Group(click.Group):
    """
    Subclass of click.Group, mostly just for supporting command aliases
    """

    def list_commands(self, ctx):
        """
        Override  from super.

        This just re-orders aliases to the bottom of --help
        """
        result = click.Group.list_commands(self, ctx)
        root_commands = [x for x in result if not self.commands[x].is_alias]
        alias_commands = [x for x in result if self.commands[x].is_alias]
        result = root_commands + alias_commands
        return result

    def command(self, *_args, **kwargs):
        """
        Override  from super.
        """
        aliases = kwargs.pop("aliases", [])

        def decorator(f):
            cmd = click.decorators.command(*_args, **kwargs)(click.pass_context(f))
            self.add_command(cmd)
            cmd.is_alias = False
            for alias in aliases:
                zz = kwargs.copy()

                def g(*aa, **kk):
                    return f(*aa, **kk)

                g.__doc__ = g.help = f"ALIAS for `{cmd.name}`"
                zz.update(help=f"ALIAS for `{cmd.name}`")
                alias_cmd = click.decorators.command(**zz)
                alias_cmd = alias_cmd(click.pass_context(g))
                alias_cmd.params = cmd.params
                self.add_command(alias_cmd, name=alias)
                alias_cmd.is_alias = True
            return cmd

        return decorator



"""
See the docs here:
  https://github.com/elo-enterprises/aws-secrets
"""


import functools

import click

from ssm import api, cli, util
from ssm.cli.wrapper import ApiWrapper

LOGGER = util.get_logger(__name__)


@click.command(cls=cli.Group)
def entry(*args, **kargs):  # noqa
    """
    Tool for accessing secrets
    """
    # this could update global settings here
    # ctx = {}
    # ctx['verbose'] = verbose
    # for key, value in config:
    #     ctx[key] = value


ApiWrapper = functools.partial(
    ApiWrapper,
    entry=entry,
)

list = ApiWrapper(
    fxn=api.list,
    aliases=["ls"],
    extra_options=[
        cli.options.env,
        cli.args.secret_name,
    ],
)

read = ApiWrapper(
    fxn=api.read,
    aliases=["get"],
    extra_options=[
        cli.options.env,
        cli.options.cascade,
        cli.args.secret_name,
    ],
)

delete = ApiWrapper(
    fxn=api.delete,
    aliases=[
        "rm",
    ],
    extra_options=[
        cli.options.env,
        cli.args.secret_name,
        click.option(
            "--no-backup",
            help="do not create backup file",
            is_flag=True,
            default=False,
            required=False,
        ),
    ],
)

move = ApiWrapper(
    fxn=api.move,
    aliases=["mv"],
    extra_options=[
        cli.options.src_env_default,
        cli.options.dest_env_default,
        click.argument("dest_name", nargs=1),
        click.argument("src_name", nargs=1),
    ],
)
move_many = ApiWrapper(
    fxn=api.move_many,
    aliases=["mv-many", "move-path", "mv-path"],
    extra_options=[
        cli.options.src_env_default,
        cli.options.dest_env_default,
        click.argument("dest_name", nargs=1),
        click.argument("src_name", nargs=1),
    ],
)
copy = ApiWrapper(
    fxn=api.copy,
    aliases=["cp"],
    extra_options=[
        cli.options.src_env_default,
        cli.options.dest_env_default,
        click.argument("dest_name", nargs=1),
        click.argument("src_name", nargs=1),
    ],
)
update = ApiWrapper(
    fxn=api.update,
    aliases=["put", "set"],
    extra_options=[
        cli.options.env,
        cli.options.existing_file,
        click.argument("value", default="", nargs=1),
        cli.args.secret_name,
    ],
)
get_many = ApiWrapper(
    fxn=api.get_many,
    aliases=["get-path"],
    extra_options=[
        cli.options.env,
        cli.options.cascade,
        cli.options.file_format,
        cli.args.namespace,
    ],
)
put_many = ApiWrapper(
    fxn=api.put_many,
    aliases=["put-path"],
    extra_options=[
        cli.options.env,
        cli.options.file_format_yaml_default,
        cli.args.namespace,
    ],
)



""" ssm.api

See the docs here:
  https://github.com/elo-enterprises/aws-secrets
"""

import json

import yaml

from ssm import util  # chatops, keybase

from .environment import Environment

LOGGER = util.get_logger(__name__)


# @util.memoized
def _get_handle(env=None, **kwargs):
    """ """
    assert env
    env = Environment.from_profile(env) if util.is_string(env) else env
    env.logger.info("this environment will be used for retrieving secrets")
    return env.secrets


def read(secret_name, **kwargs):
    """
    get a secret
    """
    assert secret_name, f"cannot read secret_name `{secret_name}`"
    secrets = _get_handle(**kwargs)

    # print(value,  file=sys.stdout)
    try:
        return secrets[secret_name]
    except KeyError as exc:
        LOGGER.error(f"KeyError: {exc}")
        raise SystemExit(1)


def list(secret_name, format="stdout", **kwargs):
    """
    list prefixes below the given path
    """
    secrets = _get_handle(**kwargs)
    return "\n".join(secrets.under(secret_name).keys())


def get_many(namespace, format="python", **kwargs):
    """
    get many secrets from hierarchy/namespace
    """
    assert format, "output format is required"
    secrets = _get_handle(**kwargs)
    result = secrets.under(namespace)
    if format in ["python"]:
        return result
    if format in ["yaml", "yml"]:
        return yaml.dump(result)
    elif format in ["json"]:
        return json.dumps(result)
    elif format in ["env"]:
        return "\n".join(["=".join([k.split("/")[-1], v]) for k, v in result.items()])
    else:
        raise RuntimeError(f"unrecognized output format `{format}`")


def delete(secret_name, no_backup=False, **kwargs):
    """delete secret (keeping a local-backup is default)"""

    def get_backup_file(prefix):
        return ".tmp.{}".format(prefix.replace("/", "_"))

    secrets = _get_handle(**kwargs)
    parameter = read(secret_name=secret_name, **kwargs)
    if not no_backup:
        backup = get_backup_file(secret_name)
        LOGGER.debug(f"backup to: {backup}")
        with open(backup, "w") as fhandle:
            fhandle.write(parameter)
        del secrets[secret_name]
        return parameter


def move(src_name, dest_name, src_env=None, dest_env=None, **kwargs):
    """move a secret"""
    result = copy(src_name, dest_name, src_env=src_env, dest_env=dest_env, **kwargs)
    src_env = Environment.from_profile(src_env)
    del src_env.secrets[src_name]
    return result


def move_many(src_name, dest_name, **kwargs):
    """
    move a whole path of secrets
    """
    dest_name = dest_name[:-1] if dest_name.endswith("/") else dest_name
    secrets = _get_handle(**kwargs)
    names = secrets.under(src_name).keys()
    plan, results = [], []
    for name in names:
        leaf = name[len(src_name) :]
        leaf = leaf[:-1] if leaf.endswith("/") else leaf
        leaf = leaf[1:] if leaf.startswith("/") else leaf
        args = (name, "/".join([dest_name, leaf]))
        LOGGER.debug(f"mv {args} {kwargs}")
        plan.append([args, kwargs])
    LOGGER.warning(f"plan:\n{plan}")
    for args, kwargs in plan:
        results.append(move(*args, **kwargs))
    return results


def copy(src_name, dest_name, src_env=None, dest_env=None, **kwargs):
    """copy a secret"""
    # NB: mind the signature, this code is reused by `move`
    src_env = _get_handle(env=src_env)
    dest_env = _get_handle(env=dest_env)
    dest_name = dest_name or src_name
    value = src_env[src_name]
    dest_env[dest_name] = value


def update(secret_name, value, file=None, **kwargs):
    """
    put a secret
    """
    if all([not value, not file]):
        err = (
            "when `value` is not given as second arg, "
            "then `input_file` must be provided"
        )
        LOGGER.critical(err)
        raise RuntimeError(err)
    value = value or open(file).read()
    secrets = _get_handle(**kwargs)
    secrets[secret_name] = value
    return True


def put_many(secret_name, input_file=None, **kwargs):  # noqa
    """put many secrets"""
    raise NotImplementedError()
